#ifndef DMRPSOLVER_HPP_
#define DMRPSOLVER_HPP_

#include "BasicSolver.hpp"
#include "ConflictsRestarts.hpp"
#include <vector>
#include <utility>
#include <list>

const uint16_t MAX16 = (1<<16)-1;

class DeltaClause{
	Literal lit;
	Clause* cls;

public:

	DeltaClause():lit(INVALID_LIT),cls(NULL){asrt(false);};
	DeltaClause(Clause* cl):lit((*cl)[0]),cls(cl){asrt(cl->size() > 2);};
	DeltaClause(const DeltaClause& cl):lit(cl.lit),cls(cl.cls){};
	Clause* getClause()	const			{return cls;}
	Literal	getLiteral() const			{return lit;}
	void	setLiteral(Literal l)		{lit = l;}
};

typedef vector<DeltaClause>::iterator delta_iter;
typedef vector<DeltaClause>::reverse_iterator delta_riter;

class OneLevelDeltaClauses{
	bool valid;
	vector<DeltaClause> deltas;
public:
	bool 		isValid()	const	{return valid;}
	void 		invalidate()		{valid = false; deltas.clear();}
	void 		reValidate()		{asrt(deltas.empty()); valid = true;}
	OneLevelDeltaClauses()			:valid(true){};
	uint32_t 	size()		const	{return deltas.size();}
	uint32_t 	addClauseGetIndex(Clause* cls){deltas.push_back(DeltaClause(cls));return deltas.size()-1;}
	delta_iter 	deltasBegin()		{return deltas.begin();}
	delta_iter 	deltasEnd()			{return deltas.end();}
	delta_riter deltasRBegin()		{return deltas.rbegin();}
	delta_riter deltasREnd()		{return deltas.rend();}
	DeltaClause& getDeltaAt(uint idx){asrt(idx < deltas.size()); return deltas[idx];}
};

struct MakeClause{
	uint32_t index;
	OneLevelDeltaClauses* level_clauses;
	MakeClause(){}
	MakeClause(const uint32_t idx, OneLevelDeltaClauses* dcls):index(idx),level_clauses(dcls){};
	MakeClause(const MakeClause& cp):index(cp.index),level_clauses(cp.level_clauses){};
};

enum SolverType{BASIC,DMRP};

class DMRPSolver : public BasicSolver {

	SolverType last_solver_type;

	vector<OneLevelDeltaClauses* >		delta_stack; //a vector of delta clauses for each decision level
	uint32_t 							delta_stack_idx;
	vector<OneLevelDeltaClauses*>		rubbish_level_delta_clauses; //for garbage collection
	OneLevelDeltaClauses				uw_clses_for_mc; //stores all unwatched clauses as delta clauses to be considered in make count
	vector<vector<Clause*> >			hook_responses; //one entry for each var
	vector<uint16_t>					dec_var_count;

	vector<vector<MakeClause> >			var_makes;
	vector<Variable>					all_make_vars;


	struct MakeCountDescendingFunctor{
	    	DMRPSolver* outer ;
	    	MakeCountDescendingFunctor(DMRPSolver* s):outer(s){};
	      	bool operator() (const Literal& a, const Literal& b){
	      		return outer->var_makes[a.toVar().idx()].size() > outer->var_makes[b.toVar().idx()].size();	}
	};

protected:
	bool 	relinkAllUnwatchedClauses();
	bool	computeHooksEnsureBinariesSAT();

	bool    clsReLink(Clause* cls,const bool care_for_hooks);
	bool	clsPlaceValidLitToFront(Clause& cl, const uint32_t front,uint32_t& start);
	uint8_t clsValidLitsToFront(Clause& cl);

	void	resetNormalClsState();
	void 	clearAssignmentsKeepHooks(const bool care_for_responses);

	bool	checkUnwatchedClauses();

	bool	clsLinkHook(Clause& cl);
	void 	clsLinkHookNewLearnt(Clause& cl);
	void 	initHookResponses(vector<Clause*>& clauses);


	TrBool  dmrpSolveConsiderTimeAndConflicts(uint32_t timeout, uint64_t max_conflicts);
	TrBool 	dmrpSearchSatClause(Clause* cls, const uint32_t abs_max_confl );
	bool	bcpDmrpOK();
	void	dmrpLearntsGC(Clause* exception,uint32_t new_sz);

	void	restoreProbClsDelta(uint32_t dec_level);
	void	clearDeltaStack(uint32_t start_idx, const bool del);
	void 	addProbClsDelta(Clause& cl,const uint level);
	inline void incrementMakeCount(const Literal& lit, OneLevelDeltaClauses* ldcl, const uint32_t idx);
	void 	registerUnlinkedClausesForMakeCount();

	void 	updateDelta(uint32_t q_begin);
	void 	updateDeltaCheckClause(cls_iter_t& from,cls_iter_t& copy);
	bool    findNewHookResponses(vector<Clause*>& clses, vector<std::pair<Literal,Clause*> >& link_later);
	void 	unlinkUnhookedClauses(uint32_t q_begin);
	TrBool  checkAndClearBrokenClauses();
	Literal nextDeltaDecision();
	bool  	collectPossibleDecLits(vector<Literal>& possible_lits,DeltaClause& dcls);
	uint32_t computeMakeCount(Literal lit, uint32_t lower_bound, Clause* ommit);

	inline void	deltaProbClausesEnsureLevel(uint32_t level);
	void	clearDeltaClausesAndMakeCount();
	void	removeRubbishDeltaClauses(const bool clear_global);
	bool	removeRubbishMakeCounts(Variable var);

	inline void incDecCount(const Literal& lit);
	inline uint16_t getDecCount(const Literal& lit);

	void 	cancelUntilCheck(uint32_t level);
	int32_t	findNewWatcherConsiderHooks(Clause& cl,const Literal& false_lit, uint32_t other_watcher_idx,int32_t& out_free_idx);

	inline vector<Clause*>& litHookResponse(const Literal& lit) ;

	void 	cancelUntilKeepHooks(uint32_t level);
	inline void debugPrintHookResponses(const Literal& lit) ;

public:
	DMRPSolver(Settings& sets, Statistics& stats):
		BasicSolver(sets,stats),last_solver_type(BASIC),delta_stack_idx(0){}
	void 		init();
	TrBool 		isSAT( uint32_t timeout);
	void		printModel();
	const char*	whoSolvedInstane(){return last_solver_type == DMRP ? "DMRPSolver": "BasicSolver";}

protected:
	inline bool	debugNumberLearntUWClauseAllRight()  ;
	inline bool	debugAtLeastTwoUnassignedLits(Clause& cl,uint level) const ;
	inline bool debugOneLitIsResponsibleForClause(Clause& cl) const ;
	inline bool debugOneLitIsHookedOrTrue(Clause& cl) const ;
	inline bool debugOneLitIsHookedForWatchedClauses(vector<Clause*>& clauses, Clause* except = NULL) const;
	inline bool debugAllHookResponsesCleared() const ;
	inline bool debugAllDecCountsZero()const;
	inline bool debugNumberHookResponsiblesAllRight() const;
	inline bool debugClsIsInDelta(Clause* cls) const;
	inline bool debugAllWatchedNonDeltaClausesHooked() const;
	inline bool debugAllClsWatched(const vector<Clause*>) const;




};


void	DMRPSolver::deltaProbClausesEnsureLevel(uint32_t level){
	delta_stack_idx = level + 1;

	if(delta_stack.size() >= delta_stack_idx ){
		delta_stack.resize(delta_stack_idx);
	}else{
		for(uint i = delta_stack.size();i < delta_stack_idx;++i){
			OneLevelDeltaClauses* dcls = new OneLevelDeltaClauses();
			delta_stack.push_back(dcls);
			//all_level_delta_clauses.push_back(dcls);
		}
	}
	asrt(delta_stack[level]->isValid());
}

vector<Clause*>& DMRPSolver::litHookResponse(const Literal& lit) {
	asrt(litIsHooked(lit));
	Variable var = lit.toVar();
	asrt(var.idx() < hook_responses.size());
	return hook_responses[var.idx()];
}

void DMRPSolver::incrementMakeCount(const Literal& lit, OneLevelDeltaClauses* ldcl, const uint32_t idx){
	asrt(!litIsHooked(lit));
	Variable var = lit.toVar();
	vector<MakeClause>& mcands = var_makes[var.idx()];
	if(mcands.empty()){all_make_vars.push_back(var);}
	mcands.push_back(MakeClause(idx,ldcl));
}



void DMRPSolver::incDecCount(const Literal& lit){
	uint16_t& count = dec_var_count[lit.toVar().idx()];
	if(count == MAX16) {
		//print_out("var reached max \n");
		for(uint i = dec_var_count.size();i--;){dec_var_count[i] >>= 6;}
	}
	++count;
}

uint16_t DMRPSolver::getDecCount(const Literal& lit){
	return dec_var_count[lit.toVar().idx()];
}




struct ClauseEqualFunc{
   	const Clause* comp;
   	ClauseEqualFunc(Clause* compare):comp(compare){asrt(compare != NULL);};
   	bool operator() (const Clause* e){return e == comp;}
};








void DMRPSolver::debugPrintHookResponses(const Literal& lit){

	DEBUG_FUNC
	vector<Clause*>& hr = litHookResponse(lit);

	//print_out("Hook responsibles of Literal %d \n",lit.toDIMACS());
	for(uint i = hr.size();i--;){
		Clause* cl = hr[i];
		cl->print(*settings.console_out);
	}
	//print_out("End \n");
}



bool	DMRPSolver::debugNumberLearntUWClauseAllRight()  {

	DEBUG_FUNC
	uint32_t learntsc = 0;
	for(uwc_iter iter 	= unwatched_clauses.begin();  iter != unwatched_clauses.end(); ++iter){
		if((*iter)->isLearnt()) ++ learntsc;
	}
	return learntsc== nbr_learnt_uw_clses;
}

bool	DMRPSolver::debugAtLeastTwoUnassignedLits(Clause& cl,uint level) const{
	DEBUG_FUNC
	uint32_t free = 0;
	for(uint32_t i = 0; i != cl.size(); ++i){
		if(litIsUnknown(cl[i]) || litAssignLevel(cl[i]) >= level){
			if(++free > 1) return true;
		}
	}
	return false;
}


bool DMRPSolver::debugAllHookResponsesCleared()const {
	DEBUG_FUNC
	for(uint i = hook_responses.size();i--;){
		if(!hook_responses[i].empty()) return false;
	}
	return true;
}


bool DMRPSolver::debugAllDecCountsZero()const {
	DEBUG_FUNC
	for(uint i = dec_var_count.size();i--;){
		if(dec_var_count[i] != 0 ) return false;
	}
	return true;
}



bool DMRPSolver::debugNumberHookResponsiblesAllRight()const{
	DEBUG_FUNC
	uint h = sum_element_sizes(hook_responses);
	uint c = numberAllLongClses() - unwatched_clauses.size();

	bool ok = h == c || (h+1) == c;
	if(!ok){
		//print_out("Sum hooks resp: %d, sum linked cls: %d\n",h,c);
	}

	return ok;
}



bool DMRPSolver::debugOneLitIsHookedForWatchedClauses(vector<Clause*>& clauses, Clause* except)const{

	DEBUG_FUNC
	for(uint32_t idx = 0;idx != clauses.size(); ++idx){
		Clause* cls = clauses[idx];
		if(cls->isRubbish() || cls == except) continue;
		if(cls->isWatched() && !debugOneLitIsHookedOrTrue(*cls)) return false;
	}
	return true;
}

bool DMRPSolver::debugClsIsInDelta(Clause* cls) const{
	DEBUG_FUNC
	for(uint outer = 0; outer != delta_stack.size();++outer){
		OneLevelDeltaClauses* d_cls = delta_stack[outer];

		for(delta_iter iter = d_cls->deltasBegin(); iter != d_cls->deltasEnd() ;++iter){
			if(iter->getClause() == cls) return true;
		}
	}
	return false;
}


bool DMRPSolver::debugAllWatchedNonDeltaClausesHooked() const{

	DEBUG_FUNC
	for(uint32_t idx = 0;idx != fix_cls.size(); ++idx){
		Clause* cls = fix_cls[idx];
		if(cls->isRubbish()) continue;
		if(cls->isWatched() && !debugClsIsInDelta(cls) && !debugOneLitIsHookedOrTrue(*cls))
				return false;
	}

	return true;
}


bool DMRPSolver::debugOneLitIsHookedOrTrue(Clause& cl) const {
	//DEBUG_FUNC
	for(uint i = 0; i != cl.size(); ++i){
		Literal lit = cl[i];
		if(litIsTrue(lit)) return true;
		if(litIsHooked(lit) && litIsNotFalse(lit)){ return true;}
	}


	//print_out("Clause is not hooked @ level %d \n",decisionLevel());
	Literal resp_cand = INVALID_LIT;

	for(uint i = 0; i != cl.size(); ++i){
		Literal lit = cl[i];

		printf("Literal %d at level %d has value %s and var hook %s \n",lit.toDIMACS(),litAssignLevel(lit),
				(litIsUnknown(lit)?"unknown":(litIsTrue(lit)?"true":"false")),
				(varHook(lit.toVar()).isUnknown()?"unknown":(varHook(lit.toVar()).isTrue()?"true":"false")));
		if(litIsHooked(lit)){
			const vector<Clause*>& hr = hook_responses[lit.toVar().idx()];
			if(find(hr,ClauseEqualFunc(&cl))){resp_cand = lit;	}
		}
	}
	printf("Clause is%s watched marked and%s learnt \n",(cl.isWatched()?"":" not"),(cl.isLearnt()?"":" not"));
	printf("Candidate was: ");
	if(resp_cand != INVALID_LIT) printf("Literal: %d \n", resp_cand.toDIMACS());
	else printf("none\n");



	return false;
}

bool DMRPSolver::debugOneLitIsResponsibleForClause(Clause& cl) const{

	DEBUG_FUNC
	for(uint i = 0; i != cl.size(); ++i){
		Literal lit = cl[i];
		if(litIsHooked(lit) && litIsNotFalse(lit)){
			const vector<Clause*>& hr = hook_responses[lit.toVar().idx()];
			if(find(hr,ClauseEqualFunc(&cl))) return true;
		}
	}
	if(settings.console_out != NULL){
		//print_out("Nobody responsible for: \n");
		cl.print(*settings.console_out);
		//print_out("Clause is%s learnt \n",(cl.isLearnt()?"":" not"));
		//print_out("Candidate was: \n");
		for(uint i = 0; i != cl.size(); ++i){
			Literal lit = cl[i];
			if(litIsHooked(lit)){
				const vector<Clause*>& hr = hook_responses[lit.toVar().idx()];
				if(find(hr,ClauseEqualFunc(&cl))){
					//print_out("Literal: %d at decision level %d \n", lit.toDIMACS(), litAssignLevel(lit));
					return false;
				}
			}
		}
		//print_out("none\n");
	}


	return false;
}


const char*  dmrp_solver_cpp_version();
#define dmrp_solver_hpp_version "DMRPSolver.hpp:     $Revision: 1.1 $"



#endif /*HYBRIDSOLVER_HPP_*/
