/*
 * Simplifier.hpp
 *
 *  Created on: April 26, 2010
 *      Author: zielke
 */

#include "Types.hpp"
#include "BasicSolver.hpp"
#include "Statistics.hpp"
#include "BitVector.hpp"

#include <vector>
#include <iostream>

#ifndef SIMPLIFIER_HPP_
#define SIMPLIFIER_HPP_

using namespace std;

class Simplifier
{

	vector<vector<Clause*> > occur;		//occur[Lit.id()] -> vector<Clause*>

	vector<TrBool> 	modelAssignment;
	vector<TrBool> 	varAssignment; 	//varAsssignment[Var.id()] -> true|false|unknown
	BitVector 		varEliminated;

	vector<Clause*> touched;
	BitVector 		isClauseTouched;

	vector<Clause*> constraints;

	//only even|odd bits of a 32 uint are set:
//	const static uint32_t evenmask = 2863311530;
//	const static uint32_t oddmask = 1431655765;

	char* simpmodelin;

	bool duringSimp;

	//parameter
	bool blockingClauses;

	bool varelim;
	bool varelimdef;
	bool varelimclause;

	uint amountOfBlockingClauses;

	uint activeClauses;
	uint varAssignments;

	//holds the IDs of the blocked clauses
	vector<uint> blockedClauseIDs;

	BasicSolver* mySolver;
	Statistics* myStats;

	uint32_t mytimeout;
	bool printAssignmentModel;

public:

	TrBool status;

	explicit Simplifier(BasicSolver* solver, Statistics* stats, bool simpblocked,bool simpvarelim, bool printModel);

	void simplify(uint32_t timeout);
	void exportConstraintsToSolver();

	void reserveSpaceForClauses(uint32_t count);
	void reserveSpaceForVars(uint32_t count);
	bool addClause(vector<Literal> lits, uint32_t id);
	bool addClause(Clause* myClause, uint32_t id);

	//getter for private data
	inline vector<Clause*> getConstraints() 	const;
	inline uint getActiveClausecount()			const;
	inline uint getVarAssignmentsCount()		const;
	inline vector<TrBool> getVarAssignments()	const;
	inline uint getAmountBlockingClauses()		const;

	inline uint32_t getLiteralCount() 				const;

	void compAssignments(vector<TrBool>& solverAssignments);
	void cleanupClauses();
private:

	void addClauseToOccur(Clause* myClause);

	bool addSetOfClausesduringElim(Variable toDel, vector<Clause*>& toAdd);
	bool subsetTest(uint32_t signature, uint32_t otherSignature);
	bool subsetTest(vector<Literal> lits, vector<Literal> otherLits);

	void findSubsumedClauses(Clause *myClause, vector<Clause*>& subsumedClauses);
	void eraseSubsumed(Clause *myClause);

	bool selfSubsumeVariant(Clause* myClause);

	bool strengthenClause(Clause* strClause, Literal toElim);

	int mergeClauses(Literal x, vector<Clause*>& pos, vector<Clause*>& neg, vector<Clause*>& merged);
	bool mayEliminate(Variable x);

	bool propagate(Literal x);


	void setSolvedStatus(TrBool status);

	void detectBlockingClauses(uint32_t timeout);
	bool detectBlockingClauses(Literal x);

	void removeClauseFromOccur(Clause* myClause, Literal occurLit);

	inline void decreaseActiveClauses();

	inline void touch(Clause* myClause);

	void deleteClause(Clause* myClause);
};


void Simplifier::decreaseActiveClauses()	{activeClauses--;
												if(activeClauses == 0)
												{
													if(duringSimp)
													{
														setSolvedStatus(TRUE_TrBool);
													}
												}
											}
void Simplifier::touch(Clause* myClause)	{
											asrt((*myClause).getID() < isClauseTouched.size());
											if(!isClauseTouched.isSet((*myClause).getID()))
											{
												isClauseTouched.setFlag((*myClause).getID());
												touched.push_back(myClause);
											}
											}


vector<Clause*> 	Simplifier::getConstraints() 					const	{return constraints;}
uint				Simplifier::getActiveClausecount()				const	{return activeClauses;}
uint				Simplifier::getVarAssignmentsCount()			const	{return varAssignments;}
vector<TrBool>		Simplifier::getVarAssignments()					const	{return varAssignment;}
uint				Simplifier::getAmountBlockingClauses()			const	{return amountOfBlockingClauses;}
uint32_t 			Simplifier::getLiteralCount()					const
{
	uint32_t count = 0;
	for(uint i=0;i<constraints.size();i++)
	{
		if(!(*constraints[i]).null())
			count += (*constraints[i]).size();
	}
	return count;
}

#endif /* SIMPLIFIER_HPP_ */
